"""
class Splitter.

Convenient tool for creating and working with folds.
"""

import random

from ._fold_storage import FoldStorage
from ._fold_storage import _FoldFile


class _Splitter(object):
    """
     Splitter needs providing some parameters to create folds and some "reader",
     that can read source.
    """
    _REST_SIZE = 100000

    def __init__(self, line_reader, column_description, seed, min_folds_count):
        self._line_reader = line_reader
        self._line_groups_ids, self._groups_ids = self._read_groups_ids()
        # line_groups_ids -- group ids of each line
        # groups_ids -- set of all group ids in dataset
        self._folds_storage = set()
        # keeps it for removing at the end of work.
        self._column_description = column_description
        self._min_folds_count = min_folds_count
        self._random = random.Random(seed)

    # line_reader -- Reader for getting lines from file. It have to support iteration through lines.
    # column_description -- The description of the features of dataset.

    def _read_groups_ids(self):
        """Find all groups in dataset and which group each line belongs."""
        line_groups_ids = []
        groups_ids = set()

        lines = self._line_reader.lines_generator()
        # Need to return group id of line and line as string.
        for group_id, _ in lines:
            line_groups_ids.append(group_id)
            groups_ids.add(group_id)
        return line_groups_ids, groups_ids

    def _make_learn_folds(self, fold_size, left_folds):
        """Prepare test sets for folds only for one permutation"""
        count_groups = len(self._groups_ids)
        if count_groups // self._min_folds_count < fold_size:
            raise AttributeError('The size of fold is too big: count_groups: {}, fold_size: {}. Const: {}'.format(
                count_groups, fold_size, self._min_folds_count)
            )

        permutation = sorted(self._groups_ids)
        self._random.shuffle(permutation)

        result = []
        current_count_folds = min(count_groups // fold_size, left_folds)
        for i in range(current_count_folds):
            result.append(set(permutation[i * fold_size: (i + 1) * fold_size]))
        return result

    def _write_folds(self, fold_storages, num, offset):
        """Learn_set contains numbers of lines. The method itself store relevant lines from dataset to fold storage."""

        generator = self._line_reader.lines_generator()
        # Need to return group id of line and line as string.

        for fold_storage in fold_storages:
            fold_storage.open()

        try:
            rest_folds = []
            rest_fold_file = self.create_fold(None, 'offset{}_rest'.format(offset), num)
            rest_fold_file.open()
            num += 1
            rest_size = 0

            for num_line, (_, line) in enumerate(generator):
                group_id = self._line_groups_ids[num_line]
                is_written = False
                for fold_storage in fold_storages:
                    if fold_storage.contains_group_id(group_id):
                        fold_storage.add(line)
                        is_written = True
                if not is_written:
                    rest_fold_file.add(line)
                    rest_size += 1
                    if rest_size >= self._REST_SIZE:
                        rest_folds.append(rest_fold_file)
                        rest_fold_file.close()
                        rest_fold_file = self.create_fold(None, 'offset{}_rest'.format(offset), num)
                        rest_fold_file.open()
                        rest_size = 0
                        num += 1

            if rest_size > 0:
                rest_fold_file.close()
                rest_folds.append(rest_fold_file)
            elif rest_fold_file.is_opened():
                rest_fold_file.close()
        finally:
            for fold_storage in fold_storages:
                fold_storage.close()

        return rest_folds

    def create_fold_sets(self, fold_size, folds_count):
        """Create all folds for all permutations."""
        folds = []
        passed_folds_count = 0

        while passed_folds_count < folds_count:
            folds.append(self._make_learn_folds(fold_size, folds_count - passed_folds_count))
            current_learn_folds = folds[-1]
            passed_folds_count += len(current_learn_folds)
        return folds

    def fold_groups_files_generator(self, folds_groups, fold_offset):
        """Create folds storages for all folds in folds_groups. Generator."""

        fold_num = 0
        for fold_group in folds_groups:
            learn_folds = []
            skipped_folds = []
            for learn_set in fold_group:
                fold_num += 1
                if fold_offset < fold_num:
                    fold_file = self.create_fold(learn_set, 'fold', fold_num)
                    learn_folds.append(fold_file)
                elif fold_offset >= fold_num:
                    fold_file = self.create_fold(learn_set, 'offset{}_skipped'.format(fold_offset), fold_num)
                    skipped_folds.append(fold_file)

            rest_folds = self._write_folds(learn_folds + skipped_folds, fold_num, fold_offset)
            yield learn_folds, skipped_folds, rest_folds

    def create_fold(self, fold_set, name, id):
        file_name = self.create_name_from_id(name, id)
        fold_file = _FoldFile(fold_set,
                              file_name,
                              sep=self._line_reader.get_separator(),
                              column_description=self._column_description)
        self._folds_storage.add(fold_file)
        return fold_file

    def clean_folds(self):
        for file in self._folds_storage:
            file.delete()

    def clean(self):
        FoldStorage.remove_dir()

    @staticmethod
    def create_name_from_id(name, id, offset=None, max_count_digits=4):
        if offset is not None:
            name = '{name}{:0>{max_count_digits}}_offset{offset}'.format(
                id,
                name=name,
                max_count_digits=max_count_digits,
                offset=offset
            )
        else:
            name = '{name}{:0>{max_count_digits}}'.format(id, name=name, max_count_digits=max_count_digits)
        return name



try:
    from xgboost.callback import TrainingCallback as XGBTrainingCallback
except:
    class XGBTrainingCallback:
        pass


from IPython.display import display

from .metrics_plotter import MetricsPlotter

class XGBPlottingCallback(XGBTrainingCallback):
    '''XGBoost callback with metrics plotting widget from CatBoost
    '''
    
    def __init__(self, total_iterations: int):
        self.plotter = None        
        self.total_iterations = total_iterations

    def after_iteration(self, model, epoch, evals_log):
        
        data_names = evals_log.keys()
        # if more than one sample is passed, consider first as train sample 
        first_train = (
            all(['valid' in data_name.lower() for data_name in data_names]) 
            and len(data_names) > 1
        )
        
        for data_name, metrics_info in evals_log.items():
            if "train" in data_name.lower() or first_train:
                train = True
                first_train = False
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                train = False
            else:
                raise Exception("Unexpected sample name during evaluation")
                
            metrics = {name: values[-1] for name, values in metrics_info.items()}
               
            if self.plotter is None:
                names = list(metrics.keys())
                self.plotter = MetricsPlotter(names, names, self.total_iterations)
                display(self.plotter._widget)
            
            self.plotter.log(epoch, train, metrics)
        
        # False to indicate training should not stop.
        return False


def lgbm_plotting_callback():
    """LightGBM callback with metrics plotting widget from CatBoost
    """
    
    plotter = None
    
    def _init(env):
        train_metrics = []
        test_metrics = []
        
        for item in env.evaluation_result_list:
            assert len(item) == 4, "Plotting was run in not suppored mode"
            data_name, eval_name = item[:2]
            
            if "train" in data_name.lower():
                train_metrics.append(eval_name)
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                test_metrics.append(eval_name)
            else:
                raise Exception("Unexpected sample name during evaluation")
             
        nonlocal plotter
        plotter = MetricsPlotter(
            train_metrics, test_metrics, env.end_iteration - env.begin_iteration)
        
        display(plotter._widget)
    
    def _callback(env):
        if plotter is None:
            _init(env)
            
        metrics = {"train": {}, "test": {}}
        
        for item in env.evaluation_result_list:
            data_name, eval_name, result = item[:3]
            
            if "train" in data_name.lower():
                metrics["train"][eval_name] = result
            elif "valid" in data_name.lower() or "test" in data_name.lower():
                metrics["test"][eval_name] = result
            else:
                raise Exception("Unexpected sample name during evaluation")
                
        plotter.log(
            env.iteration - env.begin_iteration, 
            train=True, 
            metrics=metrics["train"]
        )
        plotter.log(
            env.iteration - env.begin_iteration, 
            train=False, 
            metrics=metrics["test"]
        )
            
    _callback.order = 20
    return _callback


import warnings

from . import _catboost
fspath = _catboost.fspath


def try_plot_offline(figs):
    try:
        from plotly.offline import iplot
        from plotly.offline import init_notebook_mode
        init_notebook_mode(connected=True)
    except ImportError as e:
        warn_msg = "To draw plots you should install plotly."
        warnings.warn(warn_msg)
        raise ImportError(str(e))
    if not isinstance(figs, list):
        figs = [figs]
    for fig in figs:
        iplot(fig)


def save_plot_file(plot_file, plot_name, figs):
    warn_msg = "To draw plots you should install plotly."
    try:
        from plotly.offline import plot as plotly_plot
    except ImportError as e:
        warnings.warn(warn_msg)
        raise ImportError(str(e))

    with open(fspath(plot_file), 'w') as html_plot_file:
        html_plot_file.write('\n'.join((
            '<html>',
            '<head>',
            '<meta charset="utf-8" />',
            '<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>',
            '<title>{}</title>'.format(plot_name),
            '</head>',
            '<body>'
        )))
        for fig in figs:
            graph_div = plotly_plot(
                fig,
                output_type='div',
                show_link=False,
                include_plotlyjs=False
            )
            html_plot_file.write('\n{}\n'.format(graph_div))
        html_plot_file.write('</body>\n</html>')



from .ipythonwidget import MetricVisualizer
from .metrics_plotter import MetricsPlotter
from .callbacks import XGBPlottingCallback, lgbm_plotting_callback


def _jupyter_labextension_paths():
    return [{
        'src': 'labextension',
        'dest': 'catboost-widget',
    }]


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'nbextension',
        'dest': 'catboost-widget',
        'require': 'catboost-widget/extension'
    }]
